from __future__ import absolute_import

import redisext.models.abc


class HashMap(redisext.models.abc.Model):
    def get(self, hash_key):
        value = self.connect_to_slave().hget(self.key, hash_key)
        return self.decode(value)

    def set(self, hash_key, value):
        value = self.encode(value)
        return self.connect_to_master().hset(self.key, hash_key, value)

    def remove(self, hash_key):
        return bool(self.connect_to_master().hdel(self.key, hash_key))


class Map(redisext.models.abc.Model):
    def get(self):
        ''' Get stored value.

        :returns: stored value
        '''
        value = self.connect_to_slave().get(self.key)
        return self.decode(value)

    def set(self, value):
        ''' Store value into map.

        :param value:
        :type value:

        :returns: status if value saved or not
        :rtype: bool
        '''
        value = self.encode(value)
        return self.connect_to_master().set(self.key, value)

    def incr(self, value=1):
        ''' Increase stored value.

        :param value: value to add
        :type value: int

        :returns: current value
        '''
        value = self.connect_to_master().incr(self.key, value)
        return self.decode(value)

    def decr(self, value=1):
        ''' Decrease stored value.

        :param value: value to add
        :type value: int

        :returns: current value
        '''
        value = self.connect_to_master().decr(self.key, value)
        return self.decode(value)

    def remove(self):
        ''' Remove value from map.

        :returns: status of removal
        :rtype: bool
        '''
        return bool(self.connect_to_master().delete(self.key))

    def exists(self):
        ''' Returns status if key exists or not

        :returns: existence
        :rtype: bool
        '''
        return bool(self.connect_to_master().exists(self.key))



class IClient(object):
    ''' Interface for redis client '''

    def __init__(self, **kwargs):
        raise NotImplementedError()

    def expire(self, key, seconds):
        return self._redis.expire(key, seconds)

    def flushdb(self):
        return self._redis.flushdb()

    def delete(self, key):
        return self._redis.delete(key)

    def exists(self, key):
        return self._redis.exists(key)

    def get(self, key):
        return self._redis.get(key)

    def hdel(self, key, hash_key):
        return self._redis.hdel(key, hash_key)

    def hget(self, key, hash_key):
        return self._redis.hget(key, hash_key)

    def hset(self, key, hash_key, value):
        return self._redis.hset(key, hash_key, value)

    def llen(self, key):
        return self._redis.llen(key)

    def lpop(self, key):
        return self._redis.lpop(key)

    def lpush(self, key, item):
        return self._redis.lpush(key, item)

    def lrange(self, key, start, stop):
        return self._redis.lrange(key, start, stop)

    def ltrim(self, key, start, stop):
        return self._redis.ltrim(key, start, stop)

    def persist(self, key):
        return self._redis.persist(key)

    def rpop(self, key):
        return self._redis.rpop(key)

    def sadd(self, key, item):
        return self._redis.sadd(key, item)

    def set(self, key, value):
        return self._redis.set(key, value)

    def spop(self, key):
        return self._redis.spop(key)

    def ttl(self, key):
        return self._redis.ttl(key)

    def zadd(self, key, *args, **kwargs):
        return self._redis.zadd(key, *args, **kwargs)

    def incr(self, key, amount=1):
        return self._redis.incr(key, amount)

    def decr(self, key, amount=1):
        return self._redis.decr(key, amount)

    def zrangebyscore(self, key, min_score, max_score, start=0, num=1, withscores=False):
        return self._redis.zrangebyscore(
            key,
            min_score,
            max_score,
            start=start,
            num=num,
            withscores=withscores,
        )

    def zrem(self, key, item):
        return self._redis.zrem(key, item)

    def zcount(self, key, start_score, end_score):
        return self._redis.zcount(key, start_score, end_score)

    def zremrangebyrank(self, key, start, stop):
        return self._redis.zremrangebyrank(key, start, stop)

    def zremrangebyscore(self, key, min_score, max_score):
        return self._redis.zremrangebyscore(key, min_score, max_score)

    def zrevrange(self, key, start, stop, withscores=False):
        return self._redis.zrevrange(key, start, stop, withscores=withscores)

    def zscore(self, key, member):
        return self._redis.zscore(key, member)


class IConnection(object):
    CLIENT = IClient
    MASTER = None
    SLAVE = None

    @classmethod
    def connect_to_master(cls):
        return cls.CLIENT(**cls.MASTER)

    @classmethod
    def connect_to_slave(cls):
        if cls.SLAVE:
            connection_settings = cls.SLAVE
        else:
            connection_settings = cls.MASTER
        return cls.CLIENT(**connection_settings)



from __future__ import absolute_import

import json
import pickle


class JSON(object):
    @staticmethod
    def encode(item):
        return json.dumps(item)

    @staticmethod
    def decode(item):
        if isinstance(item, bytes):
            item = item.decode()
        return json.loads(item)


class String(object):
    @staticmethod
    def encode(item):
        return str(item)

    @staticmethod
    def decode(item):
        if isinstance(item, bytes):
            item = item.decode()
        return item


class Numeric(object):
    @staticmethod
    def encode(item):
        return int(item)

    @staticmethod
    def decode(item):
        return int(item)


class Pickle(object):
    @staticmethod
    def encode(item):
        return pickle.dumps(item)

    @staticmethod
    def decode(item):
        return pickle.loads(item)



from __future__ import absolute_import

import redisext.backend.abc
import redisext.backend.redis as redis
import redisext.packages.dsnparse as dsnparse


class Model(object):
    KEY = None
    CONNECTION = None
    CONNECTION_REUSE = True

    def __init__(self, key=None):
        self.key = key or getattr(self, 'KEY', None)

        self._connection = None
        self._master = None
        self._slave = None

    @classmethod
    def decode(cls, value):
        serializer = getattr(cls, 'SERIALIZER', None)
        if value and serializer:
            return serializer.decode(value)
        else:
            return value

    @classmethod
    def encode(cls, value):
        serializer = getattr(cls, 'SERIALIZER', None)
        if value and serializer:
            return serializer.encode(value)
        else:
            return value

    @property
    def connection(self):
        if not self._connection:
            if isinstance(self.CONNECTION, str):
                conn = dsnparse.parse(self.CONNECTION)

                class Connection(redis.Connection):
                    MASTER = {'host': conn.host, 'port': conn.port, 'db': conn.paths[0]}
                self._connection = Connection
            elif issubclass(self.CONNECTION, redisext.backend.abc.IConnection):
                self._connection = self.CONNECTION
            else:
                raise ValueError
        return self._connection

    def connect_to_master(self):
        if not self.CONNECTION_REUSE:
            return self.connection.connect_to_master()
        if not self._master:
            self._master = self.connection.connect_to_master()
        return self._master

    def connect_to_slave(self):
        if not self.CONNECTION_REUSE:
            return self.connection.connect_to_slave()
        if not self._slave:
            self._slave = self.connection.connect_to_slave()
        return self._slave



from __future__ import absolute_import

import redisext.hashmap
import redisext.key


class Lock(redisext.hashmap.Map, redisext.key.Expire):
    VALUE = ''

    def acquire(self, expiration=None):
        '''
        Acquire a lock.

        :param expiration: lock's expiration time
        :type expiration: int

        :returns: status of lock acquision
        :rtype: bool
        '''
        if self.exists():
            return False
        self.set(self.VALUE)
        if expiration:
            self.expire(expiration)
        return True

    def release(self):
        '''
        Release a lock.

        :returns: status of lock releasion
        :rtype: bool
        '''
        return self.remove()



from __future__ import absolute_import

import redisext.backend.abc
import rm.rmredis


class Client(redisext.backend.abc.IClient):
    def __init__(self, database=None, role=None):
        self._redis = rm.rmredis.RmRedis.get_instance(database, role)


class Connection(redisext.backend.abc.IConnection):
    CLIENT = Client



from __future__ import absolute_import

import redisext.backend.abc
import redis


class Client(redisext.backend.abc.IClient):
    def __init__(self, host, port, db):
        self._redis = redis.StrictRedis(host, port, db)


class Connection(redisext.backend.abc.IConnection):
    CLIENT = Client



from __future__ import absolute_import

import redisext.models.abc


class Counter(redisext.models.abc.Model):
    def get(self):
        ''' Get counter's value.

        :returns: counter's value
        :rtype: int
        '''
        value = self.connect_to_slave().get(self.key)
        return self.decode(value)

    def incr(self, value=1):
        ''' Increment counter by `value`.

        :param value: value to add
        :type value: int

        :returns: counter's value
        :rtype: int
        '''
        value = self.connect_to_master().incr(self.key, value)
        return self.decode(value)



from __future__ import absolute_import

import redisext.models.abc


class Stack(redisext.models.abc.Model):
    def pop(self):
        '''
        Pop item from stack.
        '''
        item = self.connect_to_master().lpop(self.key)
        return self.decode(item)

    def push(self, item):
        '''
        Push item into queue.

        :returns: number of items in queue
        :rtype: int
        '''
        item = self.encode(item)
        return self.connect_to_master().lpush(self.key, item)



from __future__ import absolute_import

import redisext.models.abc


class Pool(redisext.models.abc.Model):
    def pop(self):
        '''
        Pop item from pool.

        :returns: obviously item
        :rtype: how knows(serializer knows)
        '''
        item = self.connect_to_master().spop(self.key)
        return self.decode(item)

    def push(self, item):
        '''
        Place item into pool.

        :param item: whatever you need to place into pool
        :rtype: bool
        '''
        item = self.encode(item)
        return bool(self.connect_to_master().sadd(self.key, item))


class SortedSet(redisext.models.abc.Model):
    def add(self, element, score):
        element = self.encode(element)
        return bool(self.connect_to_master().zadd(self.key, score, element))

    def rem(self, element):
        element = self.encode(element)
        return bool(self.connect_to_master().zrem(self.key, element))

    def length(self, start_score, end_score):
        return int(self.connect_to_slave().zcount(self.key, start_score, end_score))

    def members(self, with_scores=False):
        elements = self.connect_to_slave().zrevrange(self.key, 0, -1, withscores=with_scores)
        if not elements:
            return elements

        if with_scores:
            return [(s, self.decode(e)) for e, s in elements]
        else:
            return [self.decode(e) for e in elements]

    def members_by_score(self, min_score='-inf', max_score='+inf', with_scores=False):
        elements = self.connect_to_slave().zrangebyscore(self.key, min_score, max_score, num=-1, withscores=with_scores)
        if not elements:
            return elements

        if with_scores:
            return [(s, self.decode(e)) for e, s in elements]
        else:
            return [self.decode(e) for e in elements]

    def contains(self, element):
        element = self.encode(element)
        return self.connect_to_slave().zscore(self.key, element) is not None

    def truncate(self, size):
        return int(self.connect_to_master().zremrangebyrank(self.key, 0, -1 * size - 1))

    def clean(self):
        return bool(self.connect_to_master().delete(self.key))

    def clean_by_score(self, min_score, max_score):
        elements = self.connect_to_master().zremrangebyscore(self.key, min_score, max_score)
        return bool(elements)



from __future__ import absolute_import

import redisext.models.abc


class Queue(redisext.models.abc.Model):
    def pop(self):
        '''
        Pop item from queue.

        :returns: item from queue
        '''
        item = self.connect_to_master().rpop(self.key)
        return self.decode(item)

    def push(self, item):
        '''
        Push item into queue.

        :param item:
        :type item:

        :returns: number of items in queue
        :rtype: int
        '''
        item = self.encode(item)
        return self.connect_to_master().lpush(self.key, item)

    def size(self):
        '''
        Get queue size.

        :returns: number of items in queue
        :rtype: int
        '''
        return self.connect_to_master().llen(self.key)

    def items(self):
        '''
        Get all queue items.

        :returns: ordered list of items
        :rtype: list
        '''
        items = self.connect_to_slave().lrange(self.key, 0, -1)
        return [self.decode(i) for i in items]


class PriorityQueue(redisext.models.abc.Model):
    def pop(self):
        redis = self.connect_to_master()
        item = redis.zrangebyscore(self.key, '-inf', '+inf', num=1)
        item = item[0] if item else None
        redis.zrem(self.key, item)
        return self.decode(item)

    def push(self, item, priority):
        item = self.encode(item)
        return self.connect_to_master().zadd(self.key, int(priority), item)



from __future__ import absolute_import

import sys
import re

if sys.version_info[0] == 2:
    import urlparse
elif sys.version_info[0] == 3:
    import urllib.parse as urlparse
else:
    raise RuntimeError


def parse(dsn, **defaults):
    assert re.match('^\S+://\S+', dsn), '{} is invalid, only full dsn urls (scheme://host...) allowed'.format(dsn)

    first_colon = dsn.find(':')
    scheme = dsn[0:first_colon]
    dsn_url = dsn[first_colon+1:]
    url = urlparse.urlparse(dsn_url)

    options = {}
    if url.query:
        for k, kv in urlparse.parse_qs(url.query, True, True).items():
            if len(kv) > 1:
                options[k] = kv
            else:
                options[k] = kv[0]

    r = ParseResult(
        scheme=scheme,
        hostname=url.hostname,
        path=url.path,
        params=url.params,
        query=options,
        fragment=url.fragment,
        username=url.username,
        password=url.password,
        port=url.port,
        query_str=url.query,
    )
    for k, v in defaults.items():
        r.setdefault(k, v)

    return r


class ParseResult(object):
    '''
    Hold results of a parsed dsn.

    This is very similar to urlparse.ParseResult tuple:
    http://docs.python.org/2/library/urlparse.html#results-of-urlparse-and-urlsplit

    it exposes the following attributes --
        scheme
        schemes -- if your scheme has +'s in it, then this will contain a list of schemes split by +
        path
        paths -- the path segment split by /, so '/foo/bar' would be ['foo', 'bar']
        host -- same as hostname (I just like host better)
        hostname
        hostloc -- host:port
        username
        password
        netloc
        query -- a dict of the query string
        query_str -- the raw query string
        port
        fragment
    '''
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __iter__(self):
        mapping = ['scheme', 'netloc', 'path', 'params', 'query', 'fragment']
        for k in mapping:
            yield getattr(self, k, '')

    def __getitem__(self, index):
        index = int(index)
        mapping = {
            0: 'scheme',
            1: 'netloc',
            2: 'path',
            3: 'params',
            4: 'query',
            5: 'fragment',
        }

        return getattr(self, mapping[index], '')

    @property
    def schemes(self):
        ''' Scheme, splited by `+` signs '''
        return self.scheme.split('+')

    @property
    def netloc(self):
        ''' Returns {username}:{password}@{hostname}:{port} '''
        s = ''
        prefix = ''
        if self.username:
            s += self.username
            prefix = '@'

        if self.password:
            s += ':{}'.format(self.password)
            prefix = '@'

        s += '{}{}'.format(prefix, self.hostloc)
        return s

    @property
    def paths(self):
        ''' Path attribute split by / '''
        return [_f for _f in self.path.split('/') if _f]

    @property
    def host(self):
        return self.hostname

    @property
    def hostloc(self):
        ''' Returns {host}:{port} '''
        hostloc = self.hostname
        if self.port:
            hostloc = '{}:{}'.format(hostloc, self.port)
        return hostloc

    def setdefault(self, key, val):
        if not getattr(self, key, None):
            setattr(self, key, val)

    def geturl(self):
        ''' Returns the dsn back into url form '''
        return urlparse.urlunparse((
            self.scheme,
            self.netloc,
            self.path,
            self.params,
            self.query_str,
            self.fragment,
        ))



from __future__ import absolute_import


class Expire(object):
    EXPIRE = None

    def expire(self, seconds=None):
        if seconds is None:
            try:
                seconds = int(self.EXPIRE)
            except TypeError:
                raise ValueError(seconds)
        return self.connect_to_master().expire(self.key, seconds)

    def ttl(self):
        return self.connect_to_master().ttl(self.key)

    def persist(self):
        return self.connect_to_master().persist(self.key)
